/*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TOS.
 *
 * TOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "kernel.h"

	.text

	.globl	virtual_entry

virtual_entry:

	/* set GDT */
	lgdt gdt_descr
	movl $KERNEL_DS, %eax
	movl %eax, %ds
	movl %eax, %es
	movl %eax, %ss

#	movl $boot_stack, %esp
	addl $KERNEL_OFFSET, %ebx

#	movl $thread_a_stack+4096, %esp

	movl $boot_stack+STACK_SIZE, %esp

	pushl $vpos
	pushl $vmem
	pushl $0
	call clear
	addl $12, %esp

	pushl $free_memory_list
	call setup_free_memory
	addl $4, %esp

	jmp typed_entry

# %ebx = multiboot information
setup_free_memory:
	movl (%ebx), %ecx
	testl $(1<<6), %ecx
	jz die

	movl 48(%ebx), %ecx # mmap_addr
	movl %ecx, %esi     #
	movl 44(%ebx), %edi # mmap_length
	addl %ecx, %edi	    # end of mmap
	movl $0, %ebp       # free_mem_cell

1: # %ebp = previous free mem, %ecx = mmap
	# mmap < (mmap_addr + mmap_length)
	cmpl %edi, %ecx
	jae 5f

	# mmap.type == 1
	movl 20(%ecx), %edx # type
	cmpl $1, %edx
	jnz 4f

	# mmap.addr_high == 0
	movl 8(%ecx), %edx  # addr_high
	cmpl $0, %edx
	jnz 4f

	# mmap.length_high == 0
	movl 16(%ecx), %edx # length_high
	cmpl $0, %edx
	jnz 4f

	movl 4(%ecx), %eax  # addr_begin (phys)
	addl $KERNEL_OFFSET, %eax # addr_beign (virt)
	movl 12(%ecx), %edx # length
	addl %eax, %edx     # addr_end
	# check chunk overflow
	cmpl %eax, %edx
	ja 2f
	movl $0xffffffff, %edx
2: # %eax = chunk_begin, %edx = chunk_end
	# _end < chunk_end
	cmpl $_end, %edx
	jbe 4f

	# chunk_beign < _end ==> chunk_begin = _end
	cmpl $_end, %eax
	jae 3f
	movl $_end, %eax
3: # %eax = chunk_begin, %edx = chunk_end
	subl %eax, %edx
	# chunk_size >= 16
	cmpl $16, %edx
	jb 4f

	addl $4, %edx       # %edx = array size 
	movl %edx, 4(%eax)
	movl %ebp, 8(%eax)
	movl %eax, %edx
	addl $12, %edx
	movl %edx, (%eax)
	movl %eax, %ebp
4: # NEXT
	movl (%ecx), %eax
	addl %eax, %ecx
	addl $4, %ecx
	jmp 1b
5:
	movl %ebp, free_memory_list
	ret

	.globl die
die:
	jmp die


	.data

	.align	4
gdt_descr:
	.word	GDT_ENTRIES * 8 - 1
	.long	gdt_table

	.globl	gdt_table
	.align	PAGE_SIZE
gdt_table:
	.quad 0x0000000000000000	/*  0 : NULL */
	.quad 0x0000000000000000	/*  1 : Reserved */
	.quad 0x0000000000000000	/*  2 : Reserved */
	.quad 0x0000000000000000 	/*  3 : Reserved */

	.quad 0x0000000000000000	/*  4 : Reserved */
	.quad 0x0000000000000000	/*  5 : Reserved */
	.quad 0x0000000000000000	/*  6 : Reserved */
	.quad 0x0000000000000000	/*  7 : Reserved */

	.quad 0x00cf9a000000ffff	/*  8 : Kernel Code */
	.quad 0x00cf92000000ffff	/*  9 : Kernel Data */
	.quad 0x0000000000000000	/* 10 : Reserved */
	.quad 0x0000000000000000	/* 11 : Reserved */

	.quad 0x0000000000000000	/* 12 : Reserved */
	.quad 0x0000000000000000	/* 13 : Reserved */
	.quad 0x0000000000000000	/* 14 : Reserved */
	.quad 0x0000000000000000	/* 15 : Reserved */

	.globl vmem
vmem = 0x000b8000
